use rustls::{pki_types::ServerName, ClientConfig, RootCertStore};
use std::{
    net::{SocketAddr, ToSocketAddrs},
    sync::{
        atomic::{AtomicU64, Ordering},
        Arc,
    },
    vec,
};
use tokio::{
    io::{self, AsyncReadExt, AsyncWriteExt},
    net::TcpStream,
};

use clap::Parser;
use tokio_rustls::{client::TlsStream, TlsConnector};
use url::Url;

#[derive(Debug, Parser)]
struct Args {
    #[clap(default_value = "https://127.0.0.1:443")]
    url: Url,

    #[clap(default_value = "20")]
    threads: u64,

    #[clap(default_value = "10")]
    connections_per_thread: u64,
}

#[tokio::main]
async fn main() {
    let cli = Args::parse();

    let threads = cli.threads;
    let connections_per_thread = cli.connections_per_thread;

    if cli.url.clone().scheme() != "https" {
        eprintln!("Only HTTPS is supported.");
        std::process::exit(1);
    }

    let host = &*cli.url.host_str().unwrap().to_string().leak();
    println!("{}:{}", host, cli.url.port().unwrap_or(443));
    let host_and_port: &SocketAddr = &*Box::leak(Box::new(
        format!("{}:{}", host, cli.url.port().unwrap_or(443))
            .to_socket_addrs()
            .unwrap()
            .next()
            .unwrap(),
    ));
    let content = &*format!(
        "GET {} HTTP/1.1\r\nHost: {}\r\nUser-Agent: 
Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36 Edg/126.0.0.0\r\n\r\n",
        cli.url.path(),
        cli.url.host_str().unwrap()
    ).leak();

    println!("Target: {}", cli.url);

    let connection_count = Arc::new(AtomicU64::new(0));

    let root_store = RootCertStore {
        roots: webpki_roots::TLS_SERVER_ROOTS.into(),
    };

    let config = Arc::new(
        rustls::ClientConfig::builder()
            .with_root_certificates(root_store)
            .with_no_client_auth(),
    );

    struct Connection {
        connector: TlsConnector,
        socket: TcpStream,
        stream: TlsStream<TcpStream>,
    }

    for _ in 0..threads {
        let connection_count = Arc::clone(&connection_count);

        let config = config.clone();
        tokio::spawn(async move {
            let mut streams: Vec<TlsStream<TcpStream>> = vec![];

            for _ in 0..connections_per_thread {
                let connector = TlsConnector::from(config.clone());
                let socket = TcpStream::connect(host_and_port).await.unwrap();
                let mut stream = connector
                    .connect(ServerName::try_from(host.as_bytes()).unwrap(), socket)
                    .await
                    .unwrap();

                connection_count.fetch_add(1, Ordering::Relaxed);

                stream.write_all(content.as_bytes()).await.unwrap();

                streams.push(stream);
            }

            // println!("{} Connections created.", connections_per_thread);

            loop {
                for streams in streams.iter_mut() {
                    let mut buffer = vec![];
                    match streams.read_to_end(&mut buffer).await {
                        Ok(_) => {
                        }
                        Err(e) => {
                            eprintln!("Error: {:?}", e);
                            connection_count.fetch_sub(1, Ordering::Relaxed);
                        }
                    }
                }
            }
        });
    }

    println!("Press Ctrl+C to stop.");

    let connection_count_ = connection_count.clone();
    tokio::spawn(async move {
        let mut last_count = 0;
        loop {
            let connection_count = connection_count_.load(Ordering::Relaxed);
            print!(
                "{} Connections {}/sec",
                connection_count,
                connection_count - last_count
            );
            last_count = connection_count;
            use std::io::Write;
            std::io::stdout().flush().unwrap();
            std::thread::sleep(std::time::Duration::from_secs(1));
            print!("\r");
        }
    });

    println!("{} Threads created.", threads);

    loop {
        std::thread::sleep(std::time::Duration::from_secs(1));
    }
}
